/****************************************************************************
Copyright (c) 2014-2015 Martell Malone <martellmalone@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
****************************************************************************/

// Tested on mingw-w64 with
// CFLAGS = -DCOBJMACROS
// LDFLAGS = `pkg-config --libs libvlc` -ld3d11 -ldxgi -ld3dcompiler_46

#include <vlc/vlc.h>
#include <windows.h>
#include <d3d11.h>
#include <dxgi.h>
#include <d3dcompiler.h>

// mingw-w64 branch 4.x will have fixes for this i.e I added the missing IID's
#ifdef DEFINE_GUID
#undef DEFINE_GUID
#endif
#define DEFINE_GUID(id, a, b, c, d, e, f, g, h, i, j, k) const GUID id = { a,b,c, { d,e,f,g,h,i,j,k } };

DEFINE_GUID(IID_ID3D11Texture2D,0x6f15aaf2,0xd208,0x4e89,0x9a,0xb4,0x48,0x95,0x35,0xd3,0x4f,0x9c);
DEFINE_GUID(IID_IDXGIFactory,0x7b7166ec,0x21c7,0x44ae,0xb2,0x1a,0xc9,0xae,0x32,0x1a,0xe3,0x69);
DEFINE_GUID(IID_IDXGIDevice,0x54ec77fa,0x1377,0x44e6,0x8c,0x32,0x88,0xfd,0x5f,0x44,0xc8,0x4c);

HINSTANCE                hInst;
HWND                     hWnd;
D3D_DRIVER_TYPE          driverType;
D3D_FEATURE_LEVEL        featureLevel;
ID3D11Device*            pDevice;
ID3D11DeviceContext*     pDeviceContext;
IDXGISwapChain*          pSwapChain;
ID3D11RenderTargetView*  pRenderTargetView;
ID3D11Texture2D*         pDepthStencil;
ID3D11DepthStencilView*  pDepthStencilView;
ID3D11VertexShader*      pVertexShader;
ID3D11PixelShader*       pPixelShader;
ID3D11InputLayout*       pVertexLayout;
ID3D11Buffer*            pVertexBuffer;
ID3D11Buffer*            pIndexBuffer;

// the video texture
ID3D11Texture2D* ptexture;
ID3D11ShaderResourceView *resView;
ID3D11SamplerState* sampState;

#define WIDTH 640
#define HEIGHT 480

#define VIDEOWIDTH 320
#define VIDEOHEIGHT 240

unsigned char *pixeldata[VIDEOWIDTH*VIDEOHEIGHT*4];


HRESULT InitWindow( HINSTANCE hInstance, int nCmdShow );
HRESULT InitDevice();
void CleanupDevice();
LRESULT CALLBACK    WndProc( HWND, UINT, WPARAM, LPARAM );
void Render();

static void *lock(void *data, void **p_pixels) {

  *p_pixels = pixeldata;
  return NULL;
}

static void unlock(void *data, void *id, void *const *p_pixels) {

}

static void display(void *data, void *id) {
  //(void) data;
}

int main(int argc, char *argv[]) {

  libvlc_instance_t *libvlc;
  libvlc_media_t *m;
  libvlc_media_player_t *mp;
  char const *vlc_argv[] = {
      "--no-audio",
      "--no-xlib",
  };
  int vlc_argc = sizeof(vlc_argv) / sizeof(*vlc_argv);

  if(argc < 2) {
    printf("Usage: %s <filename>\n", argv[0]);
    return -1;
  }

  //create a window and other win32 callbacks here
  hInst = GetModuleHandle(NULL);
  if(FAILED(InitWindow(hInst,SW_SHOW))) return 0;

  libvlc = libvlc_new(vlc_argc, vlc_argv);
  if(!libvlc) {
    printf("LibVLC initialization failure.\n");
    return -1;
  }

  m  = libvlc_media_new_path(libvlc, argv[1]);
  mp = libvlc_media_player_new_from_media(m);
  libvlc_media_release(m);

  //setup dx 11 here
  if(FAILED(InitDevice())) {
    CleanupDevice();
    return 0;
  }

  libvlc_video_set_callbacks(mp, lock, unlock, display, NULL);
  libvlc_video_set_format(mp, "RGBA", VIDEOWIDTH, VIDEOHEIGHT, VIDEOWIDTH*4);  
  libvlc_media_player_play(mp);
  
  MSG msg;
  memset(&msg, 0, sizeof(msg));

  // put while loop here for video via dx11
  while (msg.message != WM_QUIT) {
    if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    } else {
      Render();
    }
  }

  libvlc_media_player_stop(mp);

  CleanupDevice();

  libvlc_media_player_release(mp);
  libvlc_release(libvlc);

  return (int) msg.wParam;
}

HRESULT InitWindow(HINSTANCE hInstance, int nCmdShow) {
  WNDCLASSEX wcex;
  wcex.cbSize = sizeof( WNDCLASSEX );
  wcex.style = CS_HREDRAW | CS_VREDRAW;
  wcex.lpfnWndProc = WndProc;
  wcex.cbClsExtra = 0;
  wcex.cbWndExtra = 0;
  wcex.hInstance = hInstance;
  wcex.hIcon = LoadIcon( hInstance, ( LPCTSTR )9001 );
  wcex.hCursor = LoadCursor( NULL, IDC_ARROW );
  wcex.hbrBackground = ( HBRUSH )( COLOR_WINDOW + 1 );
  wcex.lpszMenuName = NULL;
  wcex.lpszClassName = "VLCWindowClass";
  wcex.hIconSm = LoadIcon( wcex.hInstance, ( LPCTSTR )9001 );
  if( !RegisterClassEx( &wcex ) )
    return E_FAIL;

  hInst = hInstance;
  RECT rc = { 0, 0, WIDTH, HEIGHT };
  AdjustWindowRect( &rc, WS_OVERLAPPEDWINDOW, FALSE );
  hWnd = CreateWindow( "VLCWindowClass", "VLC", WS_OVERLAPPEDWINDOW,
                       CW_USEDEFAULT, CW_USEDEFAULT, rc.right - rc.left, rc.bottom - rc.top, 
                       NULL, NULL, hInstance, NULL);
  if( !hWnd )
    return E_FAIL;

  ShowWindow( hWnd, nCmdShow );
  return S_OK;
}

static const char* globShader = "\
  Texture2D shaderTexture;\
  SamplerState SampleType;\
  \
  struct VS_INPUT\
  {\
    float4 Position   : POSITION;\
    float2 Texture    : TEXCOORD0;\
  };\
  \
  struct VS_OUTPUT\
  {\
    float4 Position   : SV_POSITION;\
    float2 Texture    : TEXCOORD0;\
  };\
  \
  VS_OUTPUT VS( VS_INPUT In )\
  {\
    VS_OUTPUT Output;\
    Output.Position = float4(In.Position.xy, 0.0f, 1.0f);\
    Output.Texture = In.Texture;\
    return Output;\
  }\
  \
  float4 PS( VS_OUTPUT In ) : SV_TARGET\
  {\
    float4 textureColor;\
    textureColor = shaderTexture.Sample(SampleType, In.Texture);\
    return textureColor;\
  }\
";

HRESULT InitDevice() {

  HRESULT hr = S_OK;
  IDXGIDevice *pDXGIDevice = NULL;
  IDXGIAdapter *pAdapter = NULL;
  IDXGIFactory *pDXGIFactory = NULL;
  UINT deviceFlags = 0;
  UINT driver = 0;

  static const D3D_DRIVER_TYPE driverAttempts[] = {
    D3D_DRIVER_TYPE_HARDWARE,
    D3D_DRIVER_TYPE_WARP,
    D3D_DRIVER_TYPE_REFERENCE,
  };

  static const D3D_FEATURE_LEVEL levelAttempts[] = {
    D3D_FEATURE_LEVEL_11_0,  // Direct3D 11.0 SM 5
    D3D_FEATURE_LEVEL_10_1,  // Direct3D 10.1 SM 4
    D3D_FEATURE_LEVEL_10_0,  // Direct3D 10.0 SM 4
    D3D_FEATURE_LEVEL_9_3,   // Direct3D 9.3  SM 3
    D3D_FEATURE_LEVEL_9_2,   // Direct3D 9.2  SM 2
    D3D_FEATURE_LEVEL_9_1,   // Direct3D 9.1  SM 2
  };

  for (driver = 0; driver < ARRAYSIZE(driverAttempts); driver++) {
    hr = D3D11CreateDevice(NULL, driverAttempts[driver], NULL, deviceFlags, 
                           levelAttempts, ARRAYSIZE(levelAttempts), D3D11_SDK_VERSION, 
                           &pDevice, &featureLevel, &pDeviceContext);
    if (SUCCEEDED(hr)) break;
  }

  if(FAILED(hr)) return hr;
        hr = ID3D11Device_QueryInterface(pDevice, &IID_IDXGIDevice, (void **)&pDXGIDevice);
  if(FAILED(hr)) return hr;
        hr = IDXGIDevice_GetAdapter(pDXGIDevice, &pAdapter);
  if(FAILED(hr)) return hr;
        hr = IDXGIAdapter_GetParent(pAdapter, &IID_IDXGIFactory, (void**)&pDXGIFactory);
  if(FAILED(hr)) return hr;

  DXGI_SWAP_CHAIN_DESC swapDesc;
  memset(&swapDesc, 0, sizeof(swapDesc));

  swapDesc.SampleDesc.Count = 1;        //The Number of Multisamples per Level
  swapDesc.SampleDesc.Quality = 0;      //between 0(lowest Quality) and one lesser than pDevice->CheckMultisampleQualityLevels
  swapDesc.BufferDesc.Width = WIDTH;
  swapDesc.BufferDesc.Height = HEIGHT;
  swapDesc.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
  swapDesc.BufferDesc.RefreshRate.Numerator = 60;
  swapDesc.BufferDesc.RefreshRate.Denominator = 1;
  swapDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
  swapDesc.BufferCount = 1;
  swapDesc.OutputWindow = hWnd;
  swapDesc.Windowed = TRUE;
  hr = IDXGIFactory_CreateSwapChain(pDXGIFactory, (IUnknown*)pDevice, &swapDesc, &pSwapChain);

  if(pDXGIDevice) IDXGIDevice_Release(pDXGIDevice);
  if(pAdapter) IDXGIAdapter_Release(pAdapter);
  if(pDXGIFactory) IDXGIFactory_Release(pDXGIFactory);

  if(FAILED(hr)) return hr;

  ID3D11Texture2D* pBackBuffer = NULL;
  hr = IDXGISwapChain_GetBuffer(pSwapChain, 0, &IID_ID3D11Texture2D, ( LPVOID* )&pBackBuffer );
  if(FAILED(hr)) return hr;

  hr = ID3D11Device_CreateRenderTargetView(pDevice, (ID3D11Resource *)pBackBuffer, NULL, &pRenderTargetView );
  ID3D11Texture2D_Release(pBackBuffer);
  if(FAILED(hr)) return hr;

  D3D11_TEXTURE2D_DESC deptTexDesc;
  memset(&deptTexDesc, 0,sizeof(deptTexDesc));

  deptTexDesc.ArraySize = 1;
  deptTexDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL;
  deptTexDesc.CPUAccessFlags = 0;
  deptTexDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
  deptTexDesc.Height = (UINT)HEIGHT;
  deptTexDesc.Width = (UINT)WIDTH;
  deptTexDesc.MipLevels = 1;
  deptTexDesc.MiscFlags = 0;
  deptTexDesc.SampleDesc.Count = swapDesc.SampleDesc.Count;
  deptTexDesc.SampleDesc.Quality = swapDesc.SampleDesc.Quality;
  deptTexDesc.Usage = D3D11_USAGE_DEFAULT;

  if(pDepthStencil) ID3D11Texture2D_Release(pDepthStencil);
  hr = ID3D11Device_CreateTexture2D(pDevice, &deptTexDesc, NULL, &pDepthStencil);
  if (FAILED(hr)) return hr;

  D3D11_DEPTH_STENCIL_VIEW_DESC depthViewDesc;
  memset(&depthViewDesc, 0, sizeof(depthViewDesc));

  depthViewDesc.Format = deptTexDesc.Format;
  depthViewDesc.ViewDimension = D3D11_DSV_DIMENSION_TEXTURE2D;
  depthViewDesc.Texture2D.MipSlice = 0;

  hr = ID3D11Device_CreateDepthStencilView(pDevice, (ID3D11Resource*)pDepthStencil, &depthViewDesc, &pDepthStencilView);
  if(FAILED(hr)) return hr;

  ID3D11DeviceContext_OMSetRenderTargets(pDeviceContext, 1, &pRenderTargetView, pDepthStencilView );

  // Setup the viewport
  D3D11_VIEWPORT vp;
  vp.Width = (FLOAT)WIDTH;
  vp.Height = (FLOAT)HEIGHT;
  vp.MinDepth = 0.0f;
  vp.MaxDepth = 1.0f;
  vp.TopLeftX = 0;
  vp.TopLeftY = 0;
  ID3D11DeviceContext_RSSetViewports(pDeviceContext, 1, &vp );

  // Compile the vertex shader
  ID3DBlob* pVSBlob = NULL;
  hr = D3DCompile(globShader, strlen(globShader), NULL, NULL, NULL, "VS", "vs_5_0", 0, 0, &pVSBlob, NULL);
  if( FAILED(hr)) {
    printf("The Vertex Shader is invalid.");
    return hr;
  }

  // Create the vertex shader
  hr = ID3D11Device_CreateVertexShader(pDevice, (void *)ID3D10Blob_GetBufferPointer(pVSBlob), ID3D10Blob_GetBufferSize(pVSBlob), NULL, &pVertexShader );

  if(FAILED(hr)) {  
    ID3D11Device_Release(pVSBlob);
    return hr;
  }

  D3D11_INPUT_ELEMENT_DESC layout[] =
  {
    { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
    { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0},
  };
  UINT numElements = ARRAYSIZE( layout );

  hr = ID3D11Device_CreateInputLayout(pDevice, layout, numElements, (void *)ID3D10Blob_GetBufferPointer(pVSBlob),
                                      ID3D10Blob_GetBufferSize(pVSBlob), &pVertexLayout );
  ID3D10Blob_Release(pVSBlob);
  if(FAILED(hr)) return hr;

  ID3D11DeviceContext_IASetInputLayout(pDeviceContext, pVertexLayout );

  ID3DBlob* pPSBlob = NULL;
  hr = D3DCompile(globShader, strlen(globShader), NULL, NULL, NULL, "PS", "ps_5_0", 0, 0, &pPSBlob, NULL);
  if(FAILED(hr)) {
    printf("The Pixel Shader is invalid.");
    return hr;
  }

  hr = ID3D11Device_CreatePixelShader(pDevice, (void *)ID3D10Blob_GetBufferPointer(pPSBlob), ID3D10Blob_GetBufferSize(pPSBlob), NULL, &pPixelShader );

  if(FAILED(hr)) {  
    ID3D11Device_Release(pPSBlob);
    return hr;
  }

  #define VERTEXWIDTH 5
  #define VERTEXHEIGHT 4

  float vertices[] = {
    -1.0f, -1.0f, -1.0f, 0.0f, 1.0f,
     1.0f, -1.0f, -1.0f, 1.0f, 1.0f,
     1.0f,  1.0f, -1.0f, 1.0f, 0.0f,
    -1.0f,  1.0f, -1.0f, 0.0f, 0.0f,
  };

  D3D11_BUFFER_DESC bd;
  memset(&bd, 0, sizeof(bd));
  bd.Usage = D3D11_USAGE_DEFAULT;
  bd.ByteWidth = sizeof(float) * VERTEXWIDTH * VERTEXHEIGHT;
  bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
  bd.CPUAccessFlags = 0;

  D3D11_SUBRESOURCE_DATA InitData;
  memset(&InitData, 0, sizeof(InitData));
  InitData.pSysMem = vertices;

  hr = ID3D11Device_CreateBuffer(pDevice, &bd, &InitData, &pVertexBuffer );
  if(FAILED(hr)) return hr;

  // Set vertex buffer
  UINT stride = sizeof(float)*5;
  UINT offset = 0;
  ID3D11DeviceContext_IASetVertexBuffers(pDeviceContext, 0, 1, &pVertexBuffer, &stride, &offset );

  WORD indices[] = {
    3, 1, 0,
    2, 1, 3,
  };
 
  bd.Usage = D3D11_USAGE_DEFAULT;
  bd.ByteWidth = sizeof( WORD ) * 6;
  bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
  bd.CPUAccessFlags = 0;
  InitData.pSysMem = indices;

  hr = ID3D11Device_CreateBuffer(pDevice, &bd, &InitData, &pIndexBuffer );
  if(FAILED(hr)) return hr;
 
  // Set index buffer
  ID3D11DeviceContext_IASetIndexBuffer(pDeviceContext, pIndexBuffer, DXGI_FORMAT_R16_UINT, 0 );

  // Set primitive topology
  ID3D11DeviceContext_IASetPrimitiveTopology(pDeviceContext, D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

  D3D11_TEXTURE2D_DESC texDesc;
  memset(&texDesc, 0, sizeof(texDesc));

  texDesc.Width = VIDEOWIDTH;
  texDesc.Height = VIDEOHEIGHT;
  texDesc.MipLevels = texDesc.ArraySize = 1;
  texDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
  texDesc.SampleDesc.Count = 1;
  texDesc.Usage = D3D11_USAGE_DYNAMIC;
  texDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
  texDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
  texDesc.MiscFlags = 0;

  hr = ID3D11Device_CreateTexture2D(pDevice, &texDesc, NULL, &ptexture);
  if (FAILED(hr)) {
    printf("Failed ID3D11Device::CreateTexture2D for quadtexture");
    return -1;
  }

  D3D11_SHADER_RESOURCE_VIEW_DESC resviewDesc;
  memset(&resviewDesc, 0, sizeof(resviewDesc));
  resviewDesc.Format = texDesc.Format;
  resviewDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
  resviewDesc.Texture2D.MipLevels = texDesc.MipLevels;

  hr = ID3D11Device_CreateShaderResourceView(pDevice, (ID3D11Resource *)ptexture, &resviewDesc, &resView);

  if (FAILED(hr)) {
    if(ptexture) ID3D11Texture2D_Release(ptexture);
    printf("Failed ID3D11Device::CreateShaderResourceView");
    return -1;
  } 

  D3D11_SAMPLER_DESC sampDesc;
  memset(&sampDesc, 0, sizeof(sampDesc));
  sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
  sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
  sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
  sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
  sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
  sampDesc.MinLOD = 0;
  sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
  hr = ID3D11Device_CreateSamplerState(pDevice, &sampDesc, &sampState);
  if (FAILED(hr)) {
    if(ptexture) ID3D11Texture2D_Release(ptexture);
    printf("Failed ID3D11Device::CreateSamplerState");
    return -1;
  } 
  return S_OK;
}

void CleanupDevice() {
// todo
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
  PAINTSTRUCT ps;
  switch(message) {
    case WM_PAINT:
      BeginPaint( hWnd, &ps );
      EndPaint( hWnd, &ps );
      break;
    case WM_DESTROY:
      PostQuitMessage( 0 );
      break;
    default:
      return DefWindowProc( hWnd, message, wParam, lParam );
  }
  return 0;
}

void Render() {
  // Update texture
  D3D11_MAPPED_SUBRESOURCE source;
  ID3D11DeviceContext_Map(pDeviceContext, (ID3D11Resource*)ptexture, 0, D3D11_MAP_WRITE_DISCARD, 0, &source);
  memcpy((void*)source.pData, (void*)pixeldata, VIDEOWIDTH*VIDEOHEIGHT*4);
  ID3D11DeviceContext_Unmap(pDeviceContext, (ID3D11Resource*)ptexture, 0); 
  // Clear the back buffer 
  float ClearColor[4] = { 1.0f, 0.125f, 0.3f, 1.0f }; // red,green,blue,alpha
  ID3D11DeviceContext_ClearRenderTargetView(pDeviceContext, pRenderTargetView, ClearColor );
  ID3D11DeviceContext_ClearDepthStencilView(pDeviceContext, pDepthStencilView, D3D11_CLEAR_DEPTH, 1.0f, 0);
  // Render the quad
  ID3D11DeviceContext_VSSetShader(pDeviceContext, pVertexShader, NULL, 0);
  ID3D11DeviceContext_PSSetShader(pDeviceContext, pPixelShader, NULL, 0);
  ID3D11DeviceContext_PSSetShaderResources(pDeviceContext, 0, 1, &resView);
  ID3D11DeviceContext_PSSetSamplers(pDeviceContext, 0, 1, &sampState);
  ID3D11DeviceContext_DrawIndexed(pDeviceContext, 6, 0, 0);
  // Present the information rendered to the back buffer to the front buffer (the screen)
  IDXGISwapChain_Present(pSwapChain, 0, 0 );
}
